using System;
using UnityEngine;
using UnityEngine.UI;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using MeshVR;
using SimpleJSON;

namespace EmissiveClothing
{
    public class EmissiveClothingPlus : MVRScript
    {

        protected JSONStorableFloat _alpha;
        protected JSONStorableFloat _alphaMultiplier;
        protected JSONStorableBool _renderOriginal;
        protected JSONStorableBool _renderOriginalWithoutDecal;
        protected JSONStorableStringChooser _texPrioritizeOrder;
        protected JSONStorableBool _shaderMode;
        // protected JSONStorableUrl _loadedShaderPath = new JSONStorableUrl("shader", "");
		protected JSONStorableBool _originalColor;
		protected JSONStorableBool _colorMode;
        protected JSONStorableColor _color;

        protected JSONStorableBool _liveReload;
        protected JSONStorableBool _verbose;
        protected JSONStorableBool _filterActive;
        protected JSONStorableString _filterStore;
        protected List<JSONStorableBool> _filterToggles;
        protected List<string> _filterNames;

        private static readonly string BUNDLE_NAME = "emissiveshader.assetbundle";

        protected List<EmissiveDAZSkinWrap> _ewraps = new List<EmissiveDAZSkinWrap>();
        protected List<DAZSkinWrap> _wraps = new List<DAZSkinWrap>();

        protected List<EmissiveMaterialOptionData> _emissiveMaterialData;

        protected Shader _emissiveShader;
        protected float _lastCheck;
        
        protected bool _initScriptDone = false;
        protected bool _rebuilding = false;
        protected bool _filterDirty = false;

        protected bool _waitForOn = false;

        const string TEX_SOURCE_DECALTEX = "_DecalTex > _AlphaTex";
        const string TEX_SOURCE_ALPHATEX = "_AlphaTex > _DecalTex";
        const string TEX_SOURCE_SPECTEX = "_SpecTex > _DecalTex";
        const string TEX_SOURCE_GLOSSTEX = "_GlossTex > _DecalTex";
        const string TEX_SOURCE_BUMPMAP = "_BumpMap > _DecalTex";
        const string TEX_SOURCE_MAINTEX = "_MainTex > _DecalTex";

        public override void Init()
        {
            try {

                _initScriptDone = false;

                _filterNames = new List<string>();
                _filterToggles = new List<JSONStorableBool>();
                _emissiveMaterialData = new List<EmissiveMaterialOptionData>();

                _color = new JSONStorableColor("Color", HSVColorPicker.RGBToHSV(1f, 1f, 1f), _ => SyncMats());
                RegisterColor(_color);
                CreateColorPicker(_color, true).label = "Color";

                _alpha = new JSONStorableFloat("ColorAlpha", 1f, (float _) => { SyncMats(); }, 0f, 4f, false, true);
                RegisterFloat(_alpha);
                CreateSlider(_alpha, true).label = "Color Alpha";
                _alpha.valNoCallback = 1.2f;

                _alphaMultiplier = new JSONStorableFloat("ColorAlphaMultiplier", 1f, (float _) => { SyncMats(); }, 0f, 4f, false, true);
                RegisterFloat(_alphaMultiplier);
                var slider = CreateSlider(_alphaMultiplier, true);
                slider.label = "Alpha Multiplier";

                _renderOriginal = new JSONStorableBool("RenderOriginalMaterial", true, (bool _) => RequiresRebuild());
                RegisterBool(_renderOriginal);
                CreateToggle(_renderOriginal, false).label = "Render Original Material";

                _shaderMode = new JSONStorableBool("ShaderEmissiveMode", false, (bool _) => RequiresRebuild());
                RegisterBool(_shaderMode);
                var toggle = CreateToggle(_shaderMode, false);
                toggle.label = "Shader Emissive Mode (Sim)";
                toggle.backgroundColor = new Color(0.9f, 0.9f, 0.7f);

                _originalColor = new JSONStorableBool("UseOriginalColor", true, (bool _) => SyncMats());
                RegisterBool(_originalColor);
                CreateToggle(_originalColor, false).label = "Use Original Color";

                _colorMode = new JSONStorableBool("ColorMode", false, (bool _) => SyncMats());
                RegisterBool(_colorMode);
                toggle = CreateToggle(_colorMode, false);
                toggle.label = "Color Mode (Premultiplied)";

                _renderOriginalWithoutDecal = new JSONStorableBool("RemoveDecalFromSource", false, (bool _) => RequiresRebuild());
                RegisterBool(_renderOriginalWithoutDecal);
                toggle = CreateToggle(_renderOriginalWithoutDecal, false);
                toggle.label = "Remove Decal texture from source when 'Render Original' is active";
                toggle.height = 80;
                toggle.backgroundColor = new Color(0.9f, 0.9f, 0.7f);

                List<string> pulseMainTypes = new List<string>();
                pulseMainTypes.Add(TEX_SOURCE_DECALTEX);
                pulseMainTypes.Add(TEX_SOURCE_ALPHATEX);
                pulseMainTypes.Add(TEX_SOURCE_SPECTEX);
                pulseMainTypes.Add(TEX_SOURCE_GLOSSTEX);
                pulseMainTypes.Add(TEX_SOURCE_BUMPMAP);
                pulseMainTypes.Add(TEX_SOURCE_MAINTEX);
                _texPrioritizeOrder = new JSONStorableStringChooser("TexSourceOrder", pulseMainTypes, TEX_SOURCE_DECALTEX, "TexSourceOrder", (string type) => { RequiresRebuild(); });
                RegisterStringChooser(_texPrioritizeOrder);
                var popup = CreateScrollablePopup(_texPrioritizeOrder, false);
                popup.label = "Texture source priority";

                // RegisterUrl(_loadedShaderPath);

                CreateSpacer(false).height = 30;

                _verbose = new JSONStorableBool("Verbose", true, (bool _) => { });
                RegisterBool(_verbose);
                toggle = CreateToggle(_verbose, false);
                toggle.label = "Verbose";
                toggle.height = 80;
                toggle.backgroundColor = new Color(0.65f, 0.85f, 0.9f);

                _liveReload = new JSONStorableBool("LiveRescan", true, (bool _) => { });
                RegisterBool(_liveReload);
                toggle = CreateToggle(_liveReload, false);
                toggle.label = "Live Rescan after emissive clothing changes";
                toggle.height = 80;
                toggle.backgroundColor = new Color(0.8f, 0.9f, 0.8f);

                var btn = CreateButton("Rescan active clothes");
                btn.button.onClick.AddListener(() => RequiresRebuild());
                btn.buttonColor = new Color(0.9f, 0.9f, 0.7f);
                btn.height = 80;

                // action callable by triggers if needed to rescan clothes
                RegisterAction(new JSONStorableAction("RescanActiveClothes", () => {
                    RequiresRebuild();
                }));

                _filterActive = new JSONStorableBool("FilterOverride", false, (bool _) => {
                    UpdateFilterUI();
                    SyncFilter();
                    });
                RegisterBool(_filterActive);
                toggle = CreateToggle(_filterActive, false);
                toggle.label = "Filter Override";
                toggle.backgroundColor = new Color(0.9f, 0.9f, 0.7f);

                _filterStore = new JSONStorableString("emissiveFilterNames", "");
                RegisterString(_filterStore);

                System.Text.StringBuilder sb = new System.Text.StringBuilder();
                sb.AppendLine("<b>Render Original Material:</b>");
                sb.AppendLine("Renders the original material together with the emissive material. Different results when emissive effect is low/transparent.");
                sb.AppendLine("");
                sb.AppendLine("<b>Shader Emissive Mode (Sim):</b>");
                sb.AppendLine("Off: Recommended standard mode. Renders the emissive effect using a duplicate material to not modify built in materials in VAM. Some clothing features will not work, such as sim enabled clothing.");
                sb.AppendLine("On: The emissive effect will be available with all clothing features, such as sim enabled clothing. Renders the emissive effect by swapping the clothing materials with the emissive one. In theory, this mode breaks if other scripts also modify the material. Recommended to only use this mode when you need sim enabled clothing!");
                sb.AppendLine("");
                sb.AppendLine("<b>Use Original Color:</b>");
                sb.AppendLine("On: Use color from the material's diffuse color.");
                sb.AppendLine("Off: Uses color from the plugin.");
                sb.AppendLine("Color tints the emissive texture source! Depending on the source texture, recoloring might not work as you expect.");
                sb.AppendLine("");
                sb.AppendLine("<b>Color Mode (Premultiplied):</b>");
                sb.AppendLine("Toggles premultiplied coloring on/off when syncing materials.");
                sb.AppendLine("");
                sb.AppendLine("<b>Color Alpha:</b>");
                sb.AppendLine("The base strength of the emissive effect which can be > 1. This value is multiplied with the alpha multiplier.");
                sb.AppendLine("");
                sb.AppendLine("<b>Alpha Multiplier:</b>");
                sb.AppendLine("This value is multiplied with the color alpha value to produce the final alpha value. The alpha multiplier is intended to be updated through other triggers and scripts!");
                sb.AppendLine("");
                sb.AppendLine("<b>Remove Decal texture from source when 'Render Original' is active:</b>");
                sb.AppendLine("If the decal from the source material should be cleared when rendering the original material. Has a greater visual impact when the emissive effect is low.");
                sb.AppendLine("");
                sb.AppendLine("<b>Texture source priority:</b>");
                sb.AppendLine("Default texture used for the emissive shader is the _DecalTex (for purpose built emissive clothing items). Prioritize to use another texture source for the emissive shader here.");
                sb.AppendLine("");
                sb.AppendLine("<b>Verbose:</b>");
                sb.AppendLine("When on, log messages about certain key events, status messages and warnings in this script.");
                sb.AppendLine("Turn off to silence this script for a better user experience.");
                sb.AppendLine("");
                sb.AppendLine("<b>Live Rescan after emissive clothing changes:</b>");
                sb.AppendLine("After loading detected (which usually happens when loading clothing items or presets), any emissive clothing item that has changed will trigger 'Rescan active clothes' to be run automatically.");
                sb.AppendLine("");
                sb.AppendLine("<b>Rescan active clothes:</b>");
                sb.AppendLine("Perform a manual rescan and emissive rebuild. Recommended to do after force-reload or clothing preset change on an active emissive clothing item.");
                sb.AppendLine("");
                sb.AppendLine("<b>Filter Override:</b>");
                sb.AppendLine("Clothing materials with '(em)' in the name are automatically emissive. Enable 'Filter Override' to manually select clothing materials to be emissive. Emissive effect on non-emissive clothing materials varies depending on the selected source texture.");
                sb.AppendLine("");
                sb.AppendLine("<b>Other:</b>");
                sb.AppendLine("You can have multiple instances of this plugin but only one emissive effect will be active per clothing item!");

                JSONStorableString infoJson = new JSONStorableString("infoText", sb.ToString());
                UIDynamicTextField textField = CreateTextField(infoJson, true);
                SetTextFieldHeight(textField, 540);

                StartCoroutine(LoadShaderAndInit());
            } catch (Exception e) {
                SuperController.LogError("Exception caught: " + e);
            }
        }

        public static void SetTextFieldHeight(UIDynamicTextField textField, int height)
        {
            LayoutElement component = textField.GetComponent<LayoutElement>();
            if (component != null)
            {
                component.minHeight = height;
                component.preferredHeight = height;
            }
            textField.height = height;
        }

        //public float GetAlphaMultiplier()
        //{
        //    return _alphaMultiplier.val;
        //}
        //public void SetAlphaMultiplier(float value)
        //{
        //    if (value < 0f)
        //        value = 0f;
        //    _alphaMultiplier.val = value;
        //}

        //public bool GetUseOriginalColor()
        //{
        //    return _originalColor.val;
        //}
        //public void SetUseOriginalColor(bool value)
        //{
        //    _originalColor.val = value;
        //}

        //public Color GetColor()
        //{
        //    return _color.colorPicker.currentColor;
        //}
        //public void SetColor(Color c)
        //{
        //    _color.colorPicker.currentColor = c;
        //}

        public void RequiresRebuild(float delaySeconds = 0f)
        {
            if (!_initScriptDone || _rebuilding)
                return;

            _rebuilding = true;
            StartCoroutine(Rebuild(delaySeconds));
        }

        public void UpdateFilterUI()
        {
            // Read wraps and create filter

            // remove old UI
            foreach(JSONStorableBool b in _filterToggles)
            {
                RemoveToggle(b);
            }
            _filterToggles.Clear();

            if (!_filterActive.val)
            {
                return;
            }
            
            DAZSkinWrap[] allWraps = containingAtom.gameObject.GetComponentsInChildren<DAZSkinWrap>(false);

            List<FilterOption> availableFilters = new List<FilterOption>();
            foreach (DAZSkinWrap wrap in allWraps)
            {
                // SuperController.LogMessage("wrap=" + wrap.ToString());
                if (wrap.ToString().Contains("(" + nameof(EmissiveDAZSkinWrap) + ")"))
                {
                    continue;
                }
                if (wrap.skin.delayDisplayOneFrame)
                {
                    continue;
                }

                bool hasMaterialSlots = wrap.GetComponents<DAZSkinWrapMaterialOptions>().Any(mo => mo.paramMaterialSlots?.Length > 0);
                if (!hasMaterialSlots)
                {
                    continue;
                }

                int materialOption = 0;
                foreach (DAZSkinWrapMaterialOptions option in wrap.GetComponents<DAZSkinWrapMaterialOptions>())
                {
                    if (option.paramMaterialSlots?.Length > 0)
                    {
                        //if (option.overrideId.Contains("(em)"))
                        //{
                        //    // a purpose built emissive clothing item
                        //}

                        string key = wrap.ToString() + "_" + wrap.wrapName + "_" + option.storeId;
                        FilterOption filterOption = new FilterOption(key, GetClothingName(option.storeId), materialOption, FancifyMaterialOptionLabel(wrap.wrapName, option.storeId));
                        availableFilters.Add(filterOption);

                        // SuperController.LogMessage(key);

                    }
                    materialOption++;
                }

            }

            availableFilters.Sort((a, b) => {
                int r = string.Compare(a.Name, b.Name);
                if (r == 0)
                {
                    if (a.MaterialSlot == b.MaterialSlot)
                        return 0;
                    else
                        return (a.MaterialSlot < b.MaterialSlot ? 1 : -1);
                }

                return r;
            });

            // SuperController.LogMessage("Available filters: " + string.Join(", ", availableFiltersEm.ToArray()));
            ReadFilter();

            int bgC = 0;
            string lastName = "";

            foreach (FilterOption option in availableFilters)
            {
                if (option.Key == null || option.Key == "")
                    continue;

                if (lastName != option.Name)
                {
                    bgC++;
                    lastName = option.Name;
                }

                bool initialSelected = _filterNames.Contains(option.Key);
                JSONStorableBool toggle = new JSONStorableBool(option.Key, initialSelected, (bool _) => SyncFilter());
                var tg = CreateToggle(toggle, false);
                tg.height = 120;
                tg.backgroundColor = (bgC % 2 == 0 ? new Color(0.8f, 0.9f, 0.9f) : new Color(0.8f, 0.85f, 0.9f));
                tg.label = option.Label;
                                
                //if (tg.label.Contains("(em)"))
                //{
                //    // a purpose built emissive clothing item
                //    tg.backgroundColor = (bgC % 2 == 0 ? new Color(0.9f, 0.9f, 0.8f) : new Color(0.9f, 0.9f, 0.9f));
                //}
                
                _filterToggles.Add(toggle);
            }
        }

        protected string GetClothingName(string storeId)
        {
            storeId = storeId.Replace("(" + nameof(DAZSkinWrap) + ")", "");

            // remove author to shorten the label, the user will understand which clothing item they are working on
            int index = storeId.IndexOf(":");
            if (index > 0)
            {
                storeId = storeId.Substring(index + 1);
            }
            // Add some space around the last "Material" word
            index = storeId.LastIndexOf("Material");
            if (index > 0)
            {
                storeId = storeId.Substring(0, index);
            }
            return storeId;
        }

        protected string FancifyMaterialOptionLabel(string wrapName, string label)
        {
            
            label = label.Replace("(" + nameof(DAZSkinWrap) + ")", "");

            // remove author to shorten the label, the user will understand which clothing item they are working on
            int index = label.IndexOf(":");
            if (index > 0)
            {
                label = label.Substring(index + 1);
            }

            // Add some space around the last "Material" word
            index = label.LastIndexOf("Material");
            if (index > 0)
            {
                label = label.Insert(index + "Material".Length, " ");
                label = label.Insert(index, "\n");
            }
            
            if (!string.IsNullOrEmpty(wrapName) && wrapName != "Normal")
            {
                index = label.IndexOf("\n");
                if (index > 0)
                {
                    label = label.Insert(index, $" ({wrapName})");
                }
                else
                {
                    label = $"{label} {wrapName}";
                }
            }
            
            return label;
        }

        protected void SyncFilter()
        {
            // forbidden as long as script is initing or rebuild is already in action
            if (!_initScriptDone || _rebuilding)
                return;

            _rebuilding = true;
            if (_verbose.val)
                SuperController.LogMessage("EmissiveClothing: Syncing filters");

            if (_filterActive.val)
            {
                string filter = string.Join(";", _filterToggles.Where(x => x.val).Select(x => x.name).ToArray());
                _filterStore.val = filter;
                _filterDirty = true;
            }
            ReadFilter();

            _rebuilding = true;
            StartCoroutine(Rebuild());
        }

        protected void ReadFilter()
        {
            string filter = _filterStore.val;
            _filterNames.Clear();
            _filterNames.AddRange(filter.Split(new char[] { ';' }));
            if (filter != "" && _filterDirty)
            {
                // only log when filter is dirty to avoid "spam" as the filter is updated through different paths in the code
                if (_verbose.val)
                {
                    SuperController.LogMessage("EmissiveClothing: Filters");
                    foreach (string s in _filterNames)
                    {
                        SuperController.LogMessage(s);
                    }
                }
            }
            _filterDirty = false;
        }

        class FilterOption
        {
            public string Key;
            public string Name;
            public int MaterialSlot;
            public string Label;
            public FilterOption(string key, string name, int materialSlot, string label)
            {
                Key = key;
                Name = name;
                MaterialSlot = materialSlot;
                Label = label;
            }
        }

        protected bool _queueLiveRescan = false;
        void Update()
        {
            if (_emissiveMaterialData == null)
                return;

            if (_waitForOn)
            {
                if (!_initScriptDone || _rebuilding)
                    return;

                if (containingAtom != null && containingAtom.on)
                {
                    if (_verbose.val)
                        SuperController.LogMessage("EmissiveClothing: Containing atom turned on, rebuilding");

                    _waitForOn = false;
                    RequiresRebuild(0.1f);
                }
                return;
            }
            
            SyncMats();

            if (_initScriptDone && !_rebuilding)
            {
                if (IsLoading())
                {
                    _queueLiveRescan = true;
                } 
                else if (_queueLiveRescan)
                {
                    if (!_liveReload.val)
                    {
                        _queueLiveRescan = false;
                        return;
                    }

                    if (_verbose.val)
                        SuperController.LogMessage("EmissiveClothing: (Live Rescan) Performing check for emissive clothing item changes");

                    _queueLiveRescan = false;

                    bool rebuild = false;

                    for (int i = 0; i < _wraps.Count; i++)
                    {
                        if (CheckWrapMaterialsChanged(_wraps[i], _ewraps[i]))
                        {
                            if (_verbose.val)
                                SuperController.LogMessage("EmissiveClothing: Wraps changed, performing rebuild");

                            rebuild = true;
                            break;
                        }
                    }

                    if (!rebuild && !_filterActive.val)
                    {
                        rebuild = ExistEMWrapWithoutEmissiveActive();
                    }

                    if (rebuild)
                    {
                        RequiresRebuild(0.3f);
                    }
                }
            }
        }


        private bool _restoreFromJSONDone = false;
        private bool _restoredFromJSON = false;
        public override void RestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, JSONArray presetAtoms = null, bool setMissingToDefault = true)
        {
            // Ensure restore from JSON is only done once and not by misstake later when load-merge another plugin.
            // If the load-merge plugin matches this one and this has not yet run its RestoreFromJSON, the load-merge
            // will call RestoreFromJSON on 2 plugin instances. This one and the newly added from the load-merge.
            if (!_restoreFromJSONDone)
            {
                _restoreFromJSONDone = true;
                _restoredFromJSON = true;
                base.RestoreFromJSON(jc, restorePhysical, restoreAppearance, presetAtoms, setMissingToDefault);
                ReadFilter();
            }
        }

        /// <summary>
        /// Returns the path to the BUNDLE_NAME
        /// </summary>
        private string GetPluginBundlePath()
        {
            // SuperController.LogMessage("GetPluginPath");
            MVRScript script = this;

            // string loadDir = SuperController.singleton.currentLoadDir;
            // SuperController.LogMessage(SuperController.singleton.currentSaveDir);
            // SuperController.LogMessage(script.storeId); // the id of the plugin as generated by VAM
            string pluginId = script.storeId.Split('_')[0]; // first part is the plugin ID in the MVRPluginManager.
            MVRPluginManager man = (script.manager != null ? script.manager : (script.containingAtom.GetStorableByID("PluginManager") as MVRPluginManager));
            string pathToScriptFile = man.GetJSON(true, true)["plugins"][pluginId].Value; // the path to the plugin file source
            // SuperController.LogMessage(pathToScriptFile); // "Custom/Scripts/FolderA/ScriptFile.cs/cslist"

            if (pathToScriptFile.Contains(":/"))
            {
                // var packaged script
                string varPackageName = pathToScriptFile.Substring(0, pathToScriptFile.IndexOf(":/"));
                return $"{varPackageName}:/Custom/Assets/{BUNDLE_NAME}";
            }
            // non var packaged script
            // simply require the emissiveshader.assetbundle to be in Custom/Assets/emissiveshader.assetbundle to let VAM only load 1 instance of this assetbundle
            return $"Custom/Assets/{BUNDLE_NAME}";

            /*
            // If the emissiveshader.assetbundle should be required to be next to the file itself:
            string pathToScriptFolder = "";
            if (pathToScriptFile.Contains("/") || pathToScriptFile.Contains("\\")) { pathToScriptFolder = pathToScriptFile.Substring(0, pathToScriptFile.LastIndexOfAny(new char[] { '/', '\\' })); }
            if (pathToScriptFolder.StartsWith(".")) { pathToScriptFolder = pathToScriptFolder.Substring(1); }
            if (pathToScriptFolder.StartsWith("/")) { pathToScriptFolder = pathToScriptFolder.Substring(1); }
            // SuperController.LogMessage(pathToScriptFolder); // "Scripts/FolderA/PluginFolder" OR "AddonPackageVarFile.#:/Scripts/FolderA/PluginFolder" for var-packages.
            return pathToScriptFolder + "/emissiveshader.assetbundle";
            */
        }

        public void SyncMats()
        {
            if (_emissiveMaterialData == null || !_initScriptDone || _rebuilding)
                return;

            for (int k = 0; k < _emissiveMaterialData.Count; k++)
            {
                EmissiveMaterialOptionData data = _emissiveMaterialData[k];
                for (int i = 0; i < data.MaterialEmissive.Length; i++)
                {
                    Material m = data.MaterialEmissive[i];
                    if (m == null)
                        continue;

                    Material ms = data.MaterialSource[i];
                    if (ms == null)
                        continue;

                    if (_shaderMode.val)
                        ms.renderQueue = m.renderQueue;
                    else
                        m.renderQueue = ms.renderQueue;

                    SyncMatColor(m, ms);
                }
            }
        }

        private void SyncMatColor(Material mat, Material matOriginalColorSource)
        {
            Color c = _color.colorPicker.currentColor;
            float alphaVal = Math.Max(0f, _alpha.val * _alphaMultiplier.val);
            if (alphaVal >= 0f)
            {
                if (!_colorMode.val)
                {
                    if (_originalColor.val)
                    {
                        if (matOriginalColorSource != null)
                        {
                            c = matOriginalColorSource.GetColor("_Color");
                            if (c == Color.black)
                            {
                                c = Color.white;
                            }
                        }
                    }
                    else
                    {
                        c = _color.colorPicker.currentColor;
                    }
                    c.a = alphaVal;
                }
                else
                {
                    if (_originalColor.val)
                    {
                        if (matOriginalColorSource != null)
                        {
                            c = matOriginalColorSource.GetColor("_Color");
                            if (c == Color.black)
                            {
                                c = Color.white;
                            }
                        }
                    }
                    else
                    {
                        c = _color.colorPicker.currentColor;
                    }
                    c *= alphaVal;
                }
                mat.SetColor("_Color", c);
            }
        }

        private void SyncMatEmissive(Material matEmissive, Material matSource)
        {
            // Grab decal, if not available use the alpha texture as backup
            string sourceTex = "_DecalTex";
            var eTex = matSource.GetTexture(sourceTex);

            switch (_texPrioritizeOrder.val)
            {
                case TEX_SOURCE_DECALTEX:
                    if (eTex == null && matSource.GetTexture("_AlphaTex") != null)
                    {
                        sourceTex = "_AlphaTex";
                        eTex = matSource.GetTexture(sourceTex);
                    }
                    break;
                case TEX_SOURCE_ALPHATEX:
                    if (matSource.GetTexture("_AlphaTex") != null)
                    {
                        sourceTex = "_AlphaTex";
                        eTex = matSource.GetTexture(sourceTex);
                    }
                    break;
                case TEX_SOURCE_SPECTEX:
                    if (matSource.GetTexture("_SpecTex") != null)
                    {
                        sourceTex = "_SpecTex";
                        eTex = matSource.GetTexture(sourceTex);
                    }
                    break;
                case TEX_SOURCE_GLOSSTEX:
                    if (matSource.GetTexture("_GlossTex") != null)
                    {
                        sourceTex = "_GlossTex";
                        eTex = matSource.GetTexture(sourceTex);
                    }
                    break;
                case TEX_SOURCE_BUMPMAP:
                    if (matSource.GetTexture("_BumpMap") != null)
                    {
                        sourceTex = "_BumpMap";
                        eTex = matSource.GetTexture(sourceTex);
                    }
                    break;
                case TEX_SOURCE_MAINTEX:
                    if (matSource.GetTexture("_MainTex") != null)
                    {
                        sourceTex = "_MainTex";
                        eTex = matSource.GetTexture(sourceTex);
                    }
                    break;
            }

            matEmissive.SetTexture("_MainTex", eTex);
            matEmissive.SetColor("_Color", matSource.GetColor("_Color"));
            matEmissive.SetFloat("_AlphaAdjust", matSource.GetFloat("_AlphaAdjust"));

            // As the offset and scale is from the emissive texture source too, these properties must ALSO be restored throughout the logic!
            matEmissive.SetTextureOffset("_MainTex", matSource.GetTextureOffset(sourceTex));
            matEmissive.SetTextureScale("_MainTex", matSource.GetTextureScale(sourceTex));
        }

        // MeshVR.AssetLoader provides some nice refcounting for us.
        // The downsides are the annoying callback API, that we can't suppress error messages, and that a refcount really shouldn't be created on failure, wat
        class LoadAssetBundle : CustomYieldInstruction 
        {
            AssetLoader.AssetBundleFromFileRequest req;

            private bool done;

            public bool success => req.assetBundle != null;
            public T GetAsset<T>(string name) where T : UnityEngine.Object
            {
                return req.assetBundle?.LoadAsset<T>(name);
            }

            public LoadAssetBundle(string url)
            {
                req = new AssetLoader.AssetBundleFromFileRequest {
                    path = url,
                    callback = _ => {
                        done = true;
                        if (req.assetBundle == null)
                            AssetLoader.DoneWithAssetBundleFromFile(req.path);
                    }
                };
                AssetLoader.QueueLoadAssetBundleFromFile(req);
            }

            public override bool keepWaiting => !done;
        }

        //public static bool IsErrorLogOpen()
        //{
        //    if (SuperController.singleton.errorLogPanel != null && SuperController.singleton.errorLogPanel.gameObject != null)
        //    {
        //        return SuperController.singleton.errorLogPanel.gameObject.activeSelf;
        //    }
        //    return false;
        //}

        private IEnumerator AttemptLoadShader(string url)
        {
            if (_emissiveShader != null)
                yield break;

            // bool errorLogAlreadyOpen = IsErrorLogOpen();

            {
                Shader shader = Resources.FindObjectsOfTypeAll<Shader>().Where(s => s.name == "Custom/Alazi/ExtraEmissiveComputeBuff").FirstOrDefault();
                if (shader != null)
                {
                    if (_verbose.val)
                        SuperController.LogMessage("EmissiveClothing: Found already loaded emissiveshader assetbundle with same shader name, trying to use it.");

                    _emissiveShader = shader;
                    yield break;
                }
            }

            var request = new LoadAssetBundle(url);
            yield return StartCoroutine(request);

            if (!request.success)
            {
                if (_emissiveShader == null)
                {
                    Shader shader = Resources.FindObjectsOfTypeAll<Shader>().Where(s => s.name == "Custom/Alazi/ExtraEmissiveComputeBuff").FirstOrDefault();
                    if (shader != null)
                    {
                        if (_verbose.val)
                            SuperController.LogMessage("EmissiveClothing: Found already loaded emissiveshader assetbundle with same shader name, trying to use it.");

                        //// a shader instance already exist, try to use it
                        //if (!errorLogAlreadyOpen)
                        //    SuperController.singleton.CloseErrorLogPanel();

                        _emissiveShader = shader;
                    }
                }
                yield break;
            }

            try
            {
                _emissiveShader = request.GetAsset<Shader>("Assets/EmissiveHack.shader");
                if (_emissiveShader != null)
                {
                    // _loadedShaderPath.val = url;
                }
                else
                {
                    SuperController.LogError("Bad emissiveshader assetbundle");
                    try
                    {
                        AssetLoader.DoneWithAssetBundleFromFile(url);
                    }
                    catch
                    {
                    }
                }
            }
            catch (Exception e)
            {
                if (_verbose.val)
                    SuperController.LogMessage("EmissiveClothing: Failed to load emissiveshader assetbundle: " + e.ToString());
            }

            if (_emissiveShader == null)
            {
                Shader shader = Resources.FindObjectsOfTypeAll<Shader>().Where(s => s.name == "Custom/Alazi/ExtraEmissiveComputeBuff").FirstOrDefault();
                if (shader != null)
                {
                    if (_verbose.val)
                        SuperController.LogMessage("EmissiveClothing: Found already loaded emissiveshader assetbundle with same shader name, trying to use it.");

                    //// a shader instance already exist, try to use it
                    //if (!errorLogAlreadyOpen)
                    //    SuperController.singleton.CloseErrorLogPanel();                    

                    _emissiveShader = shader;
                }
            }
        }

        // If there are multiple copies of this script loaded from different paths, avoid failures due to trying to load the same assetbundle. 
        // This is particularly important because sometimes the bundle seems to get leaked and not unload the shader.
        // Stash the url as well so that vac saves work properly even if this happens
        /* private static readonly string STASH_NAME = "Alazi.EmissiveShaderStash";
        private void StashShaderName()
        {
            PlayerPrefs.SetString(STASH_NAME, loadedShaderPath.val);
        }
        private void LoadStashedShader()
        {
            shader = Resources.FindObjectsOfTypeAll<Shader>().Where(s => s.name == "Custom/Alazi/ExtraEmissiveComputeBuff").FirstOrDefault();
            if (shader != null) {
                loadedShaderPath.val = PlayerPrefs.GetString(STASH_NAME);
            } else {
                PlayerPrefs.DeleteKey(STASH_NAME);
            }
        } */

        protected IEnumerator LoadShaderAndInit()
        {
			// SuperController.LogMessage("LoadShaderAndInit");
            // Wait until json load is done
            yield return new WaitForEndOfFrame();
            //LoadStashedShader();
			//SuperController.LogMessage("LoadStashedShader");
            if (_emissiveShader == null) 
            {

                string pluginBundlePath = GetPluginBundlePath();
                
                if (_verbose.val)
                    SuperController.LogMessage("EmissiveClothing: Shader path: " + pluginBundlePath);
                // Loading the same asset from exactly the same source is OK
                yield return AttemptLoadShader(pluginBundlePath);

                if (_emissiveShader == null) 
                {
                    SuperController.LogError("EmissiveClothing: Failed to load shader");
                    yield break;
                }
            }

            yield return WaitLoading();
            yield return new WaitForEndOfFrame();

            // let emissive plugins with filters activate FIRST by waiting a second longer after loading, let other threads run first
            // This is to avoid collisions when multiple emissive scripts overlap. This is not a supported scenario.
            if (!_filterActive.val)
            {
                yield return new WaitForSeconds(1f);
                yield return new WaitForEndOfFrame();
                yield return new WaitForEndOfFrame();
            }

            Build();

            yield return new WaitForEndOfFrame();
            UpdateFilterUI();

            _restoreFromJSONDone = true;
            _initScriptDone = true;
        }

        protected IEnumerator Rebuild(float delaySeconds = 0f)
        {
            if (_verbose.val)
                SuperController.LogMessage("EmissiveClothing: Rebuild emissive");

            if (delaySeconds > 0f)
            {
                yield return new WaitForSeconds(delaySeconds);
                yield return new WaitForEndOfFrame();
            }

            Unbuild();

            // wait for components to be destroyed
            yield return new WaitForEndOfFrame();

            yield return WaitLoading();
            yield return new WaitForEndOfFrame();

            Build();

            // once rebuild is finished, sync filterUI
            if (_filterActive.val)
            {
                UpdateFilterUI();
            }

            _rebuilding = false;
        }

        protected IEnumerator WaitLoading()
        {
            yield return new WaitForEndOfFrame();

            bool loadMessageShown = false;
            DateTime s = DateTime.Now;
            while (IsLoading())
            {
                if (!loadMessageShown)
                {
                    loadMessageShown = true;
                    if (_verbose.val)
                        SuperController.LogMessage("EmissiveClothing: Waiting for loading to complete");
                }
                
                yield return new WaitForEndOfFrame();
            }

            // wait just a bit longer if there was any loading
            if (loadMessageShown)
            {
                yield return new WaitForSeconds(0.3f);
            }
        }

        protected bool IsLoading(bool missingLoadingIconEqualsLoading = true)
        {
            if (SuperController.singleton.isLoading)
                return true;

            Transform loadingIcon = SuperController.singleton.loadingIcon;
            
            if (loadingIcon == null || loadingIcon.gameObject == null)
                return missingLoadingIconEqualsLoading;

            if (loadingIcon.gameObject.active)
                return true;

            return false;
        }

        protected HashSet<string> GetEmissiveEnabledWraps(DAZSkinWrap[] wraps)
        {
            return new HashSet<string>(wraps.Select(x => x.ToString()).
                                             Where(x => x.EndsWith("(" + nameof(EmissiveDAZSkinWrap) + ")")).
                                             SelectMany(x => {
                                                 // Return both "ClothingName (EmissiveDAZSkinWrap)" and "ClothingName (DAZSkinWrap)" as they have the emissive effect active
                                                 return new string[] {
                                                     x,
                                                     x.Substring(0, x.Length - ("(" + nameof(EmissiveDAZSkinWrap) + ")").Length) + "(" + nameof(DAZSkinWrap) + ")" };
                                             }));
        }

        protected bool ExistEMWrapWithoutEmissiveActive()
        {
            foreach(DAZSkinWrap wrap in WrapsWithoutEmissiveDAZSkinWrap())
            {
                DAZSkinWrapMaterialOptions[] allOptions = wrap.GetComponents<DAZSkinWrapMaterialOptions>();
                bool containsBuiltInEmOption = allOptions.Any(x => x.overrideId.Contains("(em)"));
                if (containsBuiltInEmOption)
                {
                    if (_verbose.val)
                        SuperController.LogMessage("EmissiveClothing: Wrap " + wrap.ToString() + " contains (em) without emissive active yet");
                    
                    return true;
                }
            }
            return false;
        }

        protected IEnumerable<DAZSkinWrap> WrapsWithoutEmissiveDAZSkinWrap()
        {
            DAZSkinWrap[] allWraps = containingAtom.gameObject.GetComponentsInChildren<DAZSkinWrap>(false);
            HashSet<string> hasEmissiveWraps = GetEmissiveEnabledWraps(allWraps);

            foreach (DAZSkinWrap wrap in allWraps)
            {
                if (wrap.skin.delayDisplayOneFrame)
                    continue;

                if (hasEmissiveWraps.Contains(wrap.ToString()))
                    continue;

                yield return wrap;
            }
        }

        private bool _firstWrapsBuilt = false;
        protected void Build()
        {
            // only build for active atoms
            if (containingAtom == null || !containingAtom.on)
            {
                if (_verbose.val)
                    SuperController.LogMessage("EmissiveClothing: Containing atom not on, waiting for it to be on");
                _waitForOn = true;
                return;
            }

            DAZSkinWrap[] allWraps = containingAtom.gameObject.GetComponentsInChildren<DAZSkinWrap>(false);

            _wraps = new List<DAZSkinWrap>();
            _ewraps = new List<EmissiveDAZSkinWrap>();

            _emissiveMaterialData = new List<EmissiveMaterialOptionData>();

            if (allWraps.Length == 0) 
            {
                if (_verbose.val)
                    SuperController.LogMessage("EmissiveClothing: No clothes loaded");
                return;
            }

            // A clothing item already having an EmissiveDAZSkinWrap also have the base DAZSkinWrap.
            HashSet<string> hasEmissiveWraps = GetEmissiveEnabledWraps(allWraps);

            // Clothing
            // \_ Wrap
            //    |- GPUmaterials (Material[])
            //    |- DAZSkinWrapMaterialOptions
            //       |- paramMaterialSlots (int[] - array over slots used in the wrap.GPUmaterials array)

            bool allSourceBlackColors = true;
            bool foundDup = false;
            foreach (DAZSkinWrap wrap in allWraps)
            {
                // SuperController.LogMessage("wrap="+wrap.ToString());
                if (wrap.skin.delayDisplayOneFrame) 
                {
                    if (_verbose.val)
                        SuperController.LogMessage($"EmissiveClothing: {wrap} is delayed, emissive not set up to handle that, ignoring wrap.");

                    continue;
                }

                string wrapName = wrap.ToString();
                if (wrapName.EndsWith("(" + nameof(EmissiveDAZSkinWrap) + ")"))
                {
                    // This is already an EmissiveDAZSkinWrap. Ignore and to reduce verbose duplicate emissive logs.
                    continue;
                }
                bool isEmissiveDuplicate = hasEmissiveWraps.Contains(wrapName);
                //if (isEmissiveDuplicate)
                //{
                //    // This is expected when having multiple instances of the same script - it's OK
                //    SuperController.LogMessage($"EmissiveClothing: found dup {wrap}, already handled by another emissive script instance, ignoring wrap.");
                //    continue;
                //}

                HashSet<string> warnedDups = new HashSet<string>();
                
                string key = wrap.ToString() + "_" + wrap.wrapName + "_";

                DAZSkinWrapMaterialOptions[] allOptions = wrap.GetComponents<DAZSkinWrapMaterialOptions>();
                // bool containsBuiltInEmOption = allOptions.Any(x => x.overrideId.Contains("(em)"));
                DAZSkinWrapMaterialOptions[] emissiveOptions = allOptions.Where(mo => {
                    // too lazy to duplicate all the code for slots2 / simpleMaterial
                    if (mo.paramMaterialSlots?.Length == 0)
                        return false;

                    if (_filterActive.val)
                    {
                        if (!_filterNames.Contains(key + mo.storeId))
                            return false;
                    }
                    else
                    {
                        if (!mo.overrideId.Contains("(em)"))
                            return false;
                    }

                    if (isEmissiveDuplicate)
                    {
                        // This is expected when having multiple instances of the same script - it's OK
                        if (warnedDups.Add(wrapName))
                        {
                            foundDup = true;
                            if (_verbose.val)
                                SuperController.LogMessage($"EmissiveClothing: found duplicate {wrap}, already handled by another emissive script instance, ignoring wrap.");
                        }
                        return false;
                    }

                    return true;
                }).ToArray();

                if (emissiveOptions.Length > 0)
                {
                    // this wrap has some emissive to do!
                    EmissiveMaterialOptionData data = new EmissiveMaterialOptionData();

                    int length = wrap.GPUmaterials.Length;
                    data.MaterialEmissive = new Material[length];
                    data.MaterialEmissiveBackup = new Material[length];
                    data.MaterialSource = new Material[length];
                    data.MaterialBackup = new Material[length];
                    data.MaterialHidden = new Material[length];
                    data.OriginalDecalTextures = new Texture[length];
                    data.DoEmissive = new bool[length];
                    data.OriginalWrapDraw = wrap.draw;

                    foreach (DAZSkinWrapMaterialOptions mo in allOptions)
                    {
                        // too lazy to duplicate all the code for slots2 / simpleMaterial
                        if (mo.paramMaterialSlots?.Length == 0)
                            continue;

                        bool isSetAsEmissive = emissiveOptions.Contains(mo);

                        foreach (int i in mo.paramMaterialSlots)
                        {
                            Material mat = wrap.GPUmaterials[i];

                            data.MaterialSource[i] = mat;
                            data.DoEmissive[i] = false;

                            if (!isSetAsEmissive)
                                continue;

                            // set up as an emissive source
                            Material ourMatEmissive = new Material(_emissiveShader);
                            Material ourMatBackup = new Material(mat.shader);

                            ourMatEmissive.CopyPropertiesFromMaterial(mat);
                            ourMatBackup.CopyPropertiesFromMaterial(mat);

                            data.MaterialEmissive[i] = ourMatEmissive;
                            data.MaterialBackup[i] = ourMatBackup;
                            data.DoEmissive[i] = true;

                            ourMatEmissive.name = mat.name;
                            ourMatBackup.name = mat.name;

                            ourMatEmissive.renderQueue = mat.renderQueue;
                            ourMatBackup.renderQueue = mat.renderQueue;

                            SyncMatEmissive(ourMatEmissive, mat);

                            Material ourMatEmissiveBak = new Material(ourMatEmissive.shader);
                            ourMatEmissiveBak.CopyPropertiesFromMaterial(ourMatEmissive);
                            ourMatEmissiveBak.name = ourMatEmissive.name;
                            ourMatEmissiveBak.renderQueue = ourMatEmissive.renderQueue;
                            data.MaterialEmissiveBackup[i] = ourMatEmissiveBak;

                            Color sourceColor = mat.GetColor("_Color");
                            if (sourceColor.r > 0 || sourceColor.g > 0 || sourceColor.b > 0)
                            {
                                allSourceBlackColors = false;
                            }

                            SyncMatColor(ourMatEmissive, mat);

                            // Looks better for the most part when _DecalTex is not cleared! 
                            // Let user decide as this is how the original emissive worked.
                            data.OriginalDecalTextures[i] = mat.GetTexture("_DecalTex");
                            if (_renderOriginalWithoutDecal.val)
                                mat.SetTexture("_DecalTex", null);

                            data.MaterialHidden[i] = new Material(mat);
                            data.MaterialHidden[i].SetFloat("_AlphaAdjust", -1f); // Hidden Toggle Alpha set to 0
                            data.MaterialHidden[i].renderQueue = mat.renderQueue;

                            // Should the emissive be the main material in VAM.
                            data.ReplaceShaderDirectlyMode = _shaderMode.val;
                            wrap.draw = true;
                        }
                    }

                    data.Finalize();

                    wrap.BroadcastMessage("OnApplicationFocus", true);
                    // SuperController.LogMessage("wrap.gameObject.AddComponent<EmissiveDAZSkinWrap>");
                    EmissiveDAZSkinWrap ewrap = wrap.gameObject.AddComponent<EmissiveDAZSkinWrap>();
                    ewrap.CopyFrom(wrap, data, _renderOriginal);
                    // SuperController.LogMessage("wraps.Add(wrap)");

                    _emissiveMaterialData.Add(data);
                    _wraps.Add(wrap);
                    _ewraps.Add(ewrap);

                }

            }

            if (foundDup)
            {
                if (_verbose.val)
                    SuperController.LogMessage("EmissiveClothing: Restoring the scene may result in wrong emissive result (depending on script load order) as duplicates were found. Consider using filters to avoid overlap between emissive clothing items.");
            }

            if (allSourceBlackColors && _wraps.Count > 0 && !_restoredFromJSON && !_firstWrapsBuilt)
            {
                // all source colors are black and this is not restored from JSON and it's the first time to rebuilding
                // Disabling this feature, instead letting black be fully white instead. There are other ways to disable the emissive effect using triggers if really needed!
                // This feature doesn't work too well when plugin has been active for a while, making *some* emissive clothing items feel broken or this plugin broken.
                //_originalColor.val = false;
                //if (_verbose.val)
                //    SuperController.LogMessage("EmissiveClothing: Auto disabled 'Use Original Color' (original color is black)");
                
            }
            if (_wraps.Count > 0)
            {
                _firstWrapsBuilt = true;
            }

            SyncMats();

            if (_verbose.val)
                SuperController.LogMessage("EmissiveClothing: Build done");

        }

        void Unbuild()
        {
            for (int i = 0; i < _wraps.Count; i++) 
            {
                try
                {
                    DAZSkinWrap wrap = _wraps[i];
                    EmissiveDAZSkinWrap ewrap = _ewraps[i];

                    ewrap.draw = false;
                    wrap.draw = true;

                    bool destroyed = false;
                    try
                    {
                        // Unity overloads "==" to check true against null on destroyed items.
                        destroyed = (wrap == null || wrap.gameObject == null);
                    } 
                    catch (Exception e)
                    {
                        destroyed = true;
                    }

                    if (!destroyed)
                    {
                        DAZSkinWrapControl control = (wrap.gameObject == null ? null : wrap.gameObject.GetComponent<DAZSkinWrapControl>());
                        if (control && (control.wrap == null || control.wrap == ewrap))
                        {
                            control.wrap = wrap;
                        }
                    } 
                    else
                    {
                        if (_verbose.val)
                            SuperController.LogMessage("EmissiveClothing: Wrap destroyed, ignoring restore");
                    }

                    bool fullyRestored = !destroyed;

                    EmissiveMaterialOptionData data = ewrap.Data;
                    if (data != null && !destroyed)
                    {
                        for (int j = 0; j < data.MaterialBackup.Length; j++)
                        {
                            if (data.MaterialBackup[j] == null || !data.DoEmissive[j])
                                continue;

                            if (data.ReplaceShaderDirectlyMode)
                            {
                                // swap back to original material source
                                // Verify that the material is the one we swapped to - and user hasn't done a forced reload while this plugin is active.
                                // if (ReferenceEquals(wrap.GPUmaterials[j], data.MaterialEmissive[j]))
                                if (!data.CheckEmissiveMaterialChanged(wrap.GPUmaterials, j))
                                {
                                    // This works as long as the current emissive material has not been changed at all
                                    wrap.GPUmaterials[j] = data.MaterialSource[j];
                                    wrap.GPUmaterials[j].SetTexture("_MainTex", data.MaterialBackup[j].GetTexture("_MainTex"));
                                    wrap.GPUmaterials[j].SetTexture("_AlphaTex", data.MaterialBackup[j].GetTexture("_AlphaTex"));
                                    wrap.GPUmaterials[j].SetTexture("_DecalTex", data.MaterialBackup[j].GetTexture("_DecalTex"));
                                    wrap.GPUmaterials[j].SetFloat("_AlphaAdjust", data.MaterialBackup[j].GetFloat("_AlphaAdjust"));
                                    wrap.GPUmaterials[j].SetColor("_Color", data.MaterialBackup[j].GetColor("_Color"));
                                    wrap.GPUmaterials[j].SetTextureOffset("_MainTex", data.MaterialBackup[j].GetTextureOffset("_MainTex"));
                                    wrap.GPUmaterials[j].SetTextureScale("_MainTex", data.MaterialBackup[j].GetTextureScale("_MainTex"));
                                }
                                else
                                {

                                    if (_verbose.val)
                                        SuperController.LogMessage("EmissiveClothing: GPUmaterials " + j.ToString() + " changed, doing partial restore in shader emissive mod (sim). Clothing probably forced reload or changed while emissive was active.");

                                    // copy materials back from the current emissive one (as it has the correct properties but wrong shader)
                                    // wrap.GPUmaterials[j] is the one with the emissive shader atm, which has updated materials
                                    // data.MaterialSource[j] is the original material + shader
                                    data.MaterialSource[j].CopyPropertiesFromMaterial(wrap.GPUmaterials[j]);
                                    wrap.GPUmaterials[j] = data.MaterialSource[j];

                                    fullyRestored = false;
                                }
                            }
                            else
                            {
                                // Verify that the material is still in use - and user hasn't done a forced reload while this plugin is active.
                                // if (ReferenceEquals(wrap.GPUmaterials[j], data.MaterialSource[j]))
                                if (!data.CheckSourceMaterialChanged(wrap.GPUmaterials, j))
                                {
                                    wrap.GPUmaterials[j].SetTexture("_MainTex", data.MaterialBackup[j].GetTexture("_MainTex"));
                                    wrap.GPUmaterials[j].SetTexture("_AlphaTex", data.MaterialBackup[j].GetTexture("_AlphaTex"));
                                    wrap.GPUmaterials[j].SetTexture("_DecalTex", data.MaterialBackup[j].GetTexture("_DecalTex"));
                                    wrap.GPUmaterials[j].SetFloat("_AlphaAdjust", data.MaterialBackup[j].GetFloat("_AlphaAdjust"));
                                    wrap.GPUmaterials[j].SetColor("_Color", data.MaterialBackup[j].GetColor("_Color"));
                                    wrap.GPUmaterials[j].SetTextureOffset("_MainTex", data.MaterialBackup[j].GetTextureOffset("_MainTex"));
                                    wrap.GPUmaterials[j].SetTextureScale("_MainTex", data.MaterialBackup[j].GetTextureScale("_MainTex"));
                                }
                                else
                                {
                                    if (_verbose.val)
                                        SuperController.LogMessage("EmissiveClothing: GPUmaterials " + j.ToString() + " changed, ignoring restore. Clothing probably forced reload or changed while emissive was active.");

                                    // wrap.GPUmaterials[j] is the original material + shader already, with already updated materials. Don't restore anything as the user is doing changes.
                                    
                                    fullyRestored = false;
                                }
                            }
                        }
                    }

                    // restore original draw, if everything else was restored as intended (no user changing clothing material textures or force reload).
                    if (fullyRestored)
                    {
                        wrap.draw = data.OriginalWrapDraw;
                    }

                    data.Clear();
                    ewrap.Clear();

                    try
                    {
                        GameObject.Destroy(ewrap);
                    } 
                    catch (Exception e)
                    {
                        SuperController.LogError(e.ToString());
                    }
                }
                catch (Exception e)
                {
                    SuperController.LogError(e.ToString());
                }
            }

            _emissiveMaterialData.Clear();
            _wraps.Clear();
            _ewraps.Clear();
        }

        bool CheckWrapMaterialsChanged(DAZSkinWrap wrap, EmissiveDAZSkinWrap ewrap)
        {
            bool destroyed = false;
            try
            {
                // Unity overloads "==" to check true against null on destroyed items.
                destroyed = (wrap == null || wrap.gameObject == null);
            }
            catch (Exception e)
            {
                destroyed = true;
            }
            if (destroyed)
            {
                return true;
            }

            EmissiveMaterialOptionData data = ewrap.Data;
            if (data != null)
            {
                for (int j = 0; j < data.MaterialBackup.Length; j++)
                {
                    if (data.MaterialBackup[j] == null || !data.DoEmissive[j])
                        continue;

                    if (data.ReplaceShaderDirectlyMode)
                    {
                        // MaterialEmissive is the active material containing the original textures if not modified
                        if (data.CheckEmissiveMaterialChanged(wrap.GPUmaterials, j))
                        {
                            return true;
                        }
                    }
                    else
                    {
                        // MaterialSource is the original and still the active material
                        if (data.CheckSourceMaterialChanged(wrap.GPUmaterials, j))
                        {
                            return true;
                        }
                    }
                }
                
            }

            return false;
        }

        void OnDestroy()
        {
			Unbuild();
			SyncMats();

            //if (_loadedAssetBundle) 
            //{
            //    //AssetLoader.DoneWithAssetBundleFromFile(loadedShaderPath.val);
            //}
        }

        // IEnumerator waiter()
        // {
        //     SuperController.LogError("Wait for 4 seconds");
        //     //Wait for 4 seconds
        //     yield return new WaitForSeconds(4);
        //     SuperController.LogError("Wait Finished 2");
        // }

        public class EmissiveMaterialOptionData
        {
            // !All arrays here are synced in length and material index!

            // The emissive materials ("our material")
            public Material[] MaterialEmissive;
            // Material emissive backup, as it was created.
            public Material[] MaterialEmissiveBackup;
            // Stores the source materials
            public Material[] MaterialSource;
            // Stores some original data from the source that needs to be restored during unbuild.
            public Material[] MaterialBackup;
            // Hidden material to be emissive
            public Material[] MaterialHidden;
            // Keeps a reference to the original source decal texture (if there was any)
            public Texture[] OriginalDecalTextures;
            // Keeps track on which materials should be emissive and not
            public bool[] DoEmissive;

            public Material[] RenderOriginalTrue;
            public Material[] RenderOriginalFalse;
            public Material[] ProcessMesh;

            // If the shader in the native wrap is replaced (which has a few more benifits and features)
            public bool ReplaceShaderDirectlyMode = false;

            public bool OriginalWrapDraw = true;

            public void Clear()
            {
                MaterialEmissive = null;
                MaterialEmissiveBackup = null;
                MaterialSource = null;
                MaterialBackup = null;
                MaterialHidden = null;
                DoEmissive = null;
                RenderOriginalTrue = null;
                RenderOriginalFalse = null;
                ProcessMesh = null;
                OriginalDecalTextures = null;
            }
            
            public bool CheckEmissiveMaterialChanged(Material[] currentGPUMaterials, int j)
            {
                // Verify that the material is still in use - and user hasn't done a forced reload while this plugin is active.
                if (ReferenceEquals(currentGPUMaterials[j], MaterialEmissive[j]))
                {
                    // check textures relevant to emissive operations
                    if (MaterialEmissive[j].GetTexture("_MainTex") != MaterialEmissiveBackup[j].GetTexture("_MainTex") ||
                        MaterialEmissive[j].GetTexture("_AlphaTex") != MaterialEmissiveBackup[j].GetTexture("_AlphaTex") ||
                        MaterialEmissive[j].GetTexture("_DecalTex") != MaterialEmissiveBackup[j].GetTexture("_DecalTex"))
                    {
                        return true;
                    }
                }
                else
                {
                    // materials no longer matching
                    return true;
                }
                // nothing changed
                return false;
            }
            
            public bool CheckSourceMaterialChanged(Material[] currentGPUMaterials, int j)
            {
                // Verify that the material is still in use - and user hasn't done a forced reload while this plugin is active.
                if (ReferenceEquals(currentGPUMaterials[j], MaterialSource[j]))
                {
                    // check textures relevant to emissive operations
                    if (MaterialSource[j].GetTexture("_MainTex") != MaterialBackup[j].GetTexture("_MainTex") ||
                        MaterialSource[j].GetTexture("_AlphaTex") != MaterialBackup[j].GetTexture("_AlphaTex") ||
                        MaterialSource[j].GetTexture("_DecalTex") != null && MaterialSource[j].GetTexture("_DecalTex") != OriginalDecalTextures[j]) // DecalTex might be null by user choice! Don't consider a null _DecalTex as a change!
                    {
                        // textures no longer matching
                        return true;
                    }
                }
                else
                {
                    // materials no longer matching
                    return true;
                }
                // nothing changed
                return false;
            }

            public void Finalize()
            {
                RenderOriginalTrue = new Material[MaterialEmissive.Length];
                RenderOriginalFalse = new Material[MaterialEmissive.Length];
                ProcessMesh = new Material[MaterialEmissive.Length];

                for (int i = 0; i < MaterialEmissive.Length; i++)
                {
                    RenderOriginalFalse[i] = MaterialHidden[i];
                    RenderOriginalTrue[i] = MaterialSource[i];

                    if (DoEmissive[i])
                    {
                        ProcessMesh[i] = MaterialEmissive[i];
                    } 
                    else
                    {
                        if (ReplaceShaderDirectlyMode)
                        {
                            // the shader in the native wrap will handle the rendering for non-emissive materials
                            RenderOriginalTrue[i] = null;
                            RenderOriginalFalse[i] = null;
                            ProcessMesh[i] = null;
                        } 
                        else
                        {
                            // The hidden material will be rendered with the source material
                            // creating a native rendered result for non-emissive materials in a clothing item
                            RenderOriginalTrue[i] = MaterialSource[i];
                            RenderOriginalFalse[i] = MaterialSource[i];
                            ProcessMesh[i] = MaterialSource[i];
                        }
                    }
                }
            }
        }

        // The only thing we absolutely need and can't get normally from 
        // DAZSkinWrap is the ComputeBuffers; Material.GetBuffer does not exist
        public class EmissiveDAZSkinWrap : DAZSkinWrap
        {
            Dictionary<int, ComputeBuffer> _emissiveMaterialVertsBuffers = new Dictionary<int, ComputeBuffer>();
            Dictionary<int, ComputeBuffer> _emissiveMaterialNormalsBuffers = new Dictionary<int, ComputeBuffer>();
            Dictionary<int, ComputeBuffer> _emissiveMaterialTangentsBuffers = new Dictionary<int, ComputeBuffer>();

            JSONStorableBool _renderOriginal;
            bool _oldOriginal;

            // DAZSkinWrap _sourceWrap;

            public EmissiveMaterialOptionData Data;

            public void CopyFrom(DAZSkinWrap wrap, EmissiveMaterialOptionData data, JSONStorableBool renderOriginal)
            {
                base.skinTransform = wrap.skinTransform;
                base.skin = wrap.skin;
                base.dazMesh = wrap.dazMesh;
                base.GPUSkinWrapper = wrap.GPUSkinWrapper;
                base.GPUMeshCompute = wrap.GPUMeshCompute;
                base.GPUmaterials = wrap.GPUmaterials;
                base.morphCopyFromTransform = wrap.morphCopyFromTransform;
                base.morphCopyFrom = wrap.morphCopyFrom;
                base.CopyMaterials();

                Data = data;

                // this._sourceWrap = wrap;
                this._renderOriginal = renderOriginal;

                base.wrapName = wrap.wrapName;
                base.wrapStore = wrap.wrapStore;

                base.surfaceOffset = wrap.surfaceOffset;
                base.defaultSurfaceOffset = wrap.defaultSurfaceOffset;
                base.additionalThicknessMultiplier = wrap.additionalThicknessMultiplier;
                base.defaultAdditionalThicknessMultiplier = wrap.defaultAdditionalThicknessMultiplier;

                if (data.ReplaceShaderDirectlyMode)
                {
                    base.draw = false;
                    // Swap the materials from the original source, works out much better!
                    for (int i = 0; i < data.DoEmissive.Length; i++)
                    {
                        if (data.DoEmissive[i])
                        {
                            wrap.GPUmaterials[i] = data.MaterialEmissive[i];
                        }
                    }
                }
                else
                {
                    wrap.draw = false;
                    base.draw = true;

                    var control = GetComponent<DAZSkinWrapControl>();
                    if (control && (control.wrap == null || control.wrap == wrap))
                    {
                        control.wrap = this;
                    }
                }
                
            }


            void LateUpdate()
            {
                if (_renderOriginal.val != _oldOriginal)
                {
                    materialVertsBuffers.Clear();
                    materialNormalsBuffers.Clear();
                    materialTangentsBuffers.Clear();
                    _oldOriginal = _renderOriginal.val;
                }

                if (Data.ReplaceShaderDirectlyMode)
                {
                    if (_renderOriginal.val)
                    {
                        // Render original in ReplaceShaderDirectlyMode mode is a bit hacky...

                        // Grab the magic buffer data from the ProcessMesh materials populated within VAM already
                        // copy it into the original shader (which isn't the active in ReplaceShaderDirectlyMode).
                        // Now we can render the correct mesh.
                        // However, emissive have some material specific overrides so that needs to be restored too for a correct look!
                        for (int i = 0; i < Data.MaterialSource.Length; i++)
                        {
                            if (Data.DoEmissive[i])
                            {
                                Data.RenderOriginalTrue[i].CopyPropertiesFromMaterial(Data.ProcessMesh[i]);
                                Data.RenderOriginalTrue[i].SetTexture("_MainTex", Data.MaterialBackup[i].GetTexture("_MainTex"));
                                Data.RenderOriginalTrue[i].SetTexture("_DecalTex", Data.MaterialBackup[i].GetTexture("_DecalTex"));
                                Data.RenderOriginalTrue[i].SetFloat("_AlphaAdjust", Data.MaterialBackup[i].GetFloat("_AlphaAdjust"));
                                Data.RenderOriginalTrue[i].SetColor("_Color", Data.MaterialBackup[i].GetColor("_Color"));
                                Data.RenderOriginalTrue[i].SetTextureOffset("_MainTex", Data.MaterialBackup[i].GetTextureOffset("_MainTex"));
                                Data.RenderOriginalTrue[i].SetTextureScale("_MainTex", Data.MaterialBackup[i].GetTextureScale("_MainTex"));
                            }
                        }

                        Material[] tempGPUMaterials = base.GPUmaterials;

                        base.GPUmaterials = Data.RenderOriginalTrue;
                        UpdateVertsGPU();
                        DrawMeshGPU();
                        
                        base.GPUmaterials = tempGPUMaterials;
                    }
                    return;
                }

                {
                    Material[] tempGPUMaterials = base.GPUmaterials;
                    if (!_renderOriginal.val)
                    {
                        base.GPUmaterials = Data.RenderOriginalFalse;
                        UpdateVertsGPU();
                        DrawMeshGPU();
                    }
                    else
                    {
                        base.GPUmaterials = Data.RenderOriginalTrue;
                        UpdateVertsGPU();
                        DrawMeshGPU();
                    }
                    base.GPUmaterials = tempGPUMaterials;
                }

                for (int i = 0; i < Data.ProcessMesh.Length; i++)
                {
                    if (Data.ProcessMesh[i] == null || !Data.DoEmissive[i])
                        continue;

                    Data.ProcessMesh[i].renderQueue = Data.MaterialSource[i].renderQueue;

                    // SuperController.LogMessage("_ClothVertices HasBuffer: " + Data.MaterialSource[i].HasBuffer("_ClothVertices").ToString());

                    UpdateBuffer(Data.ProcessMesh, _emissiveMaterialVertsBuffers, base.materialVertsBuffers, "verts", i);
                    //UpdateBuffer(Data.ProcessMesh, _emissiveMaterialNormalsBuffers, base.materialNormalsBuffers, "normals", i);
                    //UpdateBuffer(Data.ProcessMesh, _emissiveMaterialTangentsBuffers, base.materialTangentsBuffers, "tangents", i);

                    //if (Data.DoEmissive[i])
                    //{    
                    Graphics.DrawMesh(base.mesh, Matrix4x4.identity, Data.ProcessMesh[i], 0, null, i, null, false, false);
                    //} 
                    //else
                    //{
                        // Original mesh and material already drawn, don't draw again
                        // Graphics.DrawMesh(base.mesh, Matrix4x4.identity, Data.ProcessMesh[i], 0, null, i, null, dazMesh.castShadows, dazMesh.receiveShadows);
                    //}
                    
                }
            }

            private void UpdateBuffer(Material[] material, Dictionary<int, ComputeBuffer> ours, Dictionary<int, ComputeBuffer> theirs, string buf, int i)
            {
                ComputeBuffer ourB, theirB;
                ours.TryGetValue(i, out ourB);
                theirs.TryGetValue(i, out theirB);
                if (ourB != theirB) {
                    material[i].SetBuffer(buf, theirB);
                    ours[i] = theirB;
                }
            }

            //public void OnDestroy()
            //{
            //    base.OnDestroy();
            //    SuperController.LogMessage("Destroy test");
            //}

            public void Clear()
            {
                Data = null;
                _renderOriginal = null;
                materialVertsBuffers.Clear();
                materialNormalsBuffers.Clear();
                materialTangentsBuffers.Clear();
                _emissiveMaterialVertsBuffers.Clear();
                _emissiveMaterialNormalsBuffers.Clear();
                _emissiveMaterialTangentsBuffers.Clear();
            }
        }
    }

}
